#####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id: formengine.tcl,v 1.10 2000/12/01 10:59:33 artur Exp $
#
# Information:
# Modul (benutzt von Form Server)
# Formengine ist zustndigt fr ein DB Formular
# Es veranlast das Formular an die WidgetFactory zu bauen
# Baut das Fenster mit Navigations Schaltflchen
# Initialisiert das DBProxys. unten
# bernimmt das Auslesen und Setzten von Formular-Dialog (Maske)
# Eigene Implementierung von Callback Handler

Class NotifyHandler
NotifyHandler instproc addNotify {notify proc} {
    [self] instvar notifyproc
    lappend notifyproc($notify) $proc
}
NotifyHandler instproc clearNotify {notify proc} {
    [self] instvar notifyproc
    if {[info exist notifyproc($notify)]} {
	if {[set id [lsearch $notifyproc($notify) $proc]]>=0} {
	    if {[llength $notifyproc($notify)]==1} {
		unset notifyproc
	    } else {
		set notifyproc($notify) [lreplace $notifyproc($notify) $id $id]
	    }
	}
    }
}
NotifyHandler instproc invokeNotify {notify args} {
    [self] instvar notifyproc
    if {[info exist notifyproc($notify)]} {
	foreach nproc $notifyproc($notify) {
	    eval $nproc $args
	}
    }
}

Class FormEngine -superclass {InputWidget NotifyHandler} -parameter {
    form_node
    sqlconnect 
    {nestedform 0}
    {oneobject 0}
    data
    label
    linkedForm
    modi
    {reftableLink {}}
    {refattribute {}}
    {border 0}
}
# reftableLink hat Aufbau {reftable refattribute2}
# form_node - tDom Node
# nestedform - ist eigebettet
# oneobject - eingebettet und spezialisierung 1:1
# data - Bei Einbettung wird fr der Primschlssel gespeichert der zu RefValue
#        biem Proxy bergeben wird
# linkedForm - wird es gelinkt
# refattribute - benutzt fr Ausblendung und bergeben zu Proxy
FormEngine metadata add attributes
FormEngine metadata attributes {label form refattribute}
#
# Prozedures designed to overwrite by children classes
# by implementing an formular API
# If you want to buil your own procedures to Formular
# cut & copy this to your trigger.tcl file 
# kl* variables are referencen on TclX Extension keyed lists
#
# Please see the file sample/trigger.tcl for using it
FormEngine instproc update_check klvalues_ref {
    return 1
}
FormEngine instproc delete_check {} {
    return 1
}
FormEngine instproc insert_check klvalues_ref {
    return 1
}
FormEngine instproc after_delete {} {
}
FormEngine instproc reload_form {} {
}
FormEngine instproc position_check {pos} {
    return 1
}
FormEngine instproc filling_form klvalues_ref {
}
# API Procedures
FormEngine instproc getKeyValue {} {
    set tupelid [[self]::myDBProxy set tupelid]
    if {[llength $tupelid]!=1} {
	error "FormEngine [self] can't get primary key Value from $tupelid"
    }
    return [lindex [lindex $tupelid 0] 1]
}
#
# Normal Proceduress
#
# act_tupel - Aktueles gezeigtes Tupel
# old_tupel - Letztes behandeltes Tupel
# refattribute - benutzt beim gelinkten Forms (die Attribute werden nicth gezeigt) und durch setdefault bestimmt
# notnull_data - bestimmt fr insert
# def_values - Default aber nich DB defaults
# linkedForm - (bool)
# oneobject - beim Childfoms (bool)
# nestedform - (bool) beim Aggregation (kein eigenes Fenster)
# reftableLink - Beim n:m links
# modi - keyedlist z.B. ({nodelete 1} {noinsert 0}}
# emptyLinkState - wenn der Vater-Link-Formulare leer ist
FormEngine instproc init args {
    [self] instvar win dialog_container form_node \
	act_tupel old_tupel tupel_count sqlconnect attrlist refattribute \
	notnull_data def_values linkedForm oneobject nestedform reftableLink \
	modi emptyLinkState
    # Nested Form sucht selbst den Vater
    # Genau so bei InheritForm
    if {$nestedform || $oneobject} {
	set linkedForm [FormEngine searchParentFormEngine [[self] info parent]]
    }
    #puts "FormEngine Init $form_node $sqlconnect"
    set mf [FormFactory [Object autoname myformfactory] -sqlconnect $sqlconnect -hidedata $refattribute]
    set dialog_container [$mf makedialog_fromdom $form_node [self]]
    # Transferiere Datenherkunft Informationen
    set attrlist [$mf set data]
    set notnull_data [$mf set notnull_data]
    set def_values [$mf set def_values]
    $mf destroy
    #puts "DATA $attrlist"
    regexp {.*::(.+)} [self] _ win
    set win .$win
    #[self] build_window
    set table [$form_node getAttribute table]
    set primarykey [$form_node getAttribute primarykey]
    # lese Modi aus
    if {![info exist modi]} {
	set modi [tdom_attr2keyl $form_node {insertonly nodelete noupdate noinsert}]
    } else {
	# Die XML Kodierte Modi kann nur strenger sein
	set omodi [tdom_attr2keyl $form_node {insertonly nodelete noupdate noinsert}]
	foreach m [keylkeys omodi] {
	    if {[keylget omodi $m]} {
		keylset modi $m 1
	    }
	}
    }
    tdom_setdefaults modi {{insertonly 0} {nodelete 0} {noupdate 0} {noinsert 0}}    
    # !! Zugriff auf globalen Object FormServer
    # Option viewonly auswerten
    # nicht elegant!
    if {$FormServer::opts(viewonly)} {
	keylset modi nodelete 1
	keylset modi noinsert 1
	keylset modi noupdate 1
	# vielleicht ein Sicherheitloch xdobry will aber keine
	# zustztliche Sicherheit-Sicht machen 
	# keylset ${fname}::modi insertonly 0
    }
    # DBProxy initialisieren
    if {$sqlconnect!=""} {
	if {$reftableLink==""} {
	    DBProxy [self]::myDBProxy \
		-table $table -primarykey $primarykey -attrlist $attrlist \
		-sqlconnect $sqlconnect -refattribute $refattribute
	} else {
	    DBProxyRT [self]::myDBProxy \
		-table $table -primarykey $primarykey -attrlist $attrlist \
		-sqlconnect $sqlconnect -refattribute $refattribute \
		-reftable [lindex $reftableLink 0] \
		-refattribute2 [lindex $reftableLink 1]
	}
	if {$reftableLink=="" && [info exist linkedForm] && [FormEngine deepinstances $linkedForm]!=""} {
	    # berprfung von Rekursiven Form Aggregation
	    #puts "linkedForm fount $linkedForm"
	    if {$nestedform && $form_node==[$linkedForm set form_node]} {
		global lan
		tk_messageBox -message $lan(message.rekforms) -type ok -icon error
		error "XML Forms corrupt"
	    }
	    if {[$sqlconnect hasfunction getinsertid]} {
		$linkedForm addNotify create "[self] createNewTupel"
	    }
	    # Hier kann man auch ein Kaskadierendes Delete Aufbauen"
	    # funktioniert nicht beim Aggregation (Ausbaufhig)
	    if {$oneobject} {
		$linkedForm addNotify delete "[self] delete"
	    }
    }
    } 
    set act_tupel 0
    set old_tupel -1
    set emptyLinkState 0
   #[self] reload
}
FormEngine instproc map {{window {}}} {
    [self] instvar form_node win dialog_container nestedform label \
	data labelwin refattribute oneobject border modi linkedForm reftableLink
    if {$window!=""} {set win $window}
    #puts "WINDOW $win [self] $oneobject $nestedform"
    global lan
    if {!$nestedform} {
	#puts "not nested [self]"
	toplevel $win
	wm withdraw $win
	set title [$form_node getAttribute name]
	if {[info exist linkedForm] && $linkedForm!=""} {
	    append title " linked [[$linkedForm set form_node] getAttribute name]"
	    if {$reftableLink!=""} {
		append title " via [lindex $reftableLink 0]"
	    }
	}
	wm title $win $title
	$dialog_container map $win.dialog
	
	menu $win.mb -tearoff 0
	menu $win.mb.file -tearoff 0
	$win.mb add cascade -label $lan(menu.file.form) -menu $win.mb.file \
	    -underline 1
	
	# list of {Name Image Command}
	pack [[self] buildNavigation $win.navigation] \
	    -side bottom -anchor s -expand yes -fill x
	pack $win.dialog -side top -expand yes -fill both
	$win configure -menu $win.mb 
	wm protocol $win WM_DELETE_WINDOW "[self] destroy"
	wm deiconify $win
    } else {
	#puts "nested [self]"
	if {[info exist label] && $label!=""} {
	    tixLabelFrame $win -label $label
	    set bwin [$win subwidget frame]
	} else {
	    if {$border==1} {
		frame $win -relief sunken -borderwidth 1
	    } else {
		frame $win
	    }
	    set bwin $win
	}
	$dialog_container map $bwin.dialog
	
	pack [[self] buildNavigation $bwin.navigation] \
	    -side bottom -anchor s -expand yes -fill x
	pack $bwin.dialog -side top -expand yes -fill both
    }
    [self] setModi $modi
}
# Sucht den Vater Form Engine (benutzt bei Aggregation und FormLinks)
# !! keine Fehlertoleranz
FormEngine proc searchParentFormEngine {object} {
    set pf_engine {}
    for {set pf_engine $object} \
	{$pf_engine!="" && $pf_engine!="::" && 
	[$pf_engine info class]!="::FormEngine" && 
	![lcontain [[$pf_engine info class] info heritage] ::FormEngine]} {
	set pf_engine [$pf_engine info parent]
    } {
    }
    return $pf_engine
}
# Bildet die Naviagationleitst und vielleicht ein Men wenn
# Kein eingebeteter Folmular
# gibt ein Frame der Naviagationsleiste zurck
# Seiten Effect bei Bilden von Mens ($win.mb.file) 
#        als bereits initialisierte Men
FormEngine instproc buildNavigation {nwin} {
    [self] instvar win nestedform labelwin oneobject navwin sqlconnect \
	usecommands menuindex modi
    global lan
    set navwin $nwin
    if {$sqlconnect==""} {
	::label $nwin -text $lan(connect.noconnectwar)
	return $nwin
    }
    frame $nwin  -bd 1 -relief ridge
    set commands {
	{first start first Alt-a} 
	{prev back prev Alt-b}
	{next forward next_tupel Alt-f}
	{last finish last Alt-e}
	{newtupel newtupel newtupel {}}
	{update filefloppy updateManual {}}
	{undo undo undo Alt-z}
	{delete delete delete {}}
	{reload reload reload Alt-r}
	{find find filterData {}}
	{orderby orderby orderbyData {}}
	{exit exit destroy {}}
    }
    if {$oneobject} {
	# Name, Bitmap, Command
	set usecommands {update undo delete reload}
    } elseif {[keylget modi insertonly]} {
	set usecommands {next undo}
    } else { 
	if {[$sqlconnect hasfunction seek]} {
	    set usecommands {first prev next last update 
		undo delete reload find orderby}
	} else {
	    set usecommands {next newtupel update undo delete 
		reload find orderby}
	}
	tixLabelEntry $nwin.label -labelside right -label "von XX" \
	    -option {
		entry.width 4
	    }
	$nwin.label subwidget entry configure -takefocus 0
	set labelwin $nwin.label
	pack $nwin.label -side left -anchor w
	if {[$sqlconnect hasfunction seek]} {
	    bind [$nwin.label subwidget entry] <Return> "[self] setPosFromEntry"
	}
    }
    if {!$nestedform} {
	lappend usecommands {exit}
    }
    foreach command $commands {
	set name [lindex $command 0]
	if {![lcontain $usecommands $name]} { continue }
	set image [lindex $command 1]
	set execute [lindex $command 2]
	set accel [lindex $command 3]
	button $nwin.$name -image [tix getimage $image] \
	    -command "[self] $execute" \
	    -takefocus 0 -bg #c0c0c0 -highlightbackground #c0c0c0
	if {!$nestedform} {
	    if {$accel==""} {
		$win.mb.file add command -label $lan(navigation.${name}) \
		    -command "[self] $execute"
	    } else {
		$win.mb.file add command -label $lan(navigation.${name}) \
		    -command "[self] $execute" -accelerator $accel
		bind [winfo toplevel $nwin] <$accel> "[self] $execute"
	    }
	    keylset menuindex $name [$win.mb.file index $lan(navigation.${name})]
	}
	pack $nwin.$name -side left -anchor w
    }
    if {[winfo exist $nwin.next]} {
	$nwin.next configure -takefocus 1
    }
    return $nwin
}
# Setzen von nomal/disabled fr verschiedene Funktionnen
# die Prozedur erkennt welche funktionnen mglich sind
# und merk sich den internen status (commStates) 
# um zu viele neukonfiguarationen
# zu vermeiden
# mode = { normal | disabled }
FormEngine instproc setFunctionMode {mode funclist} {
    [self] instvar usecommands commStates navwin nestedform win menuindex
    if {![info exist commStates]} {
	foreach comm $usecommands {
	    keylset commStates $comm normal
	}
    }
    global lan
    foreach comm [schnittMenge $funclist $usecommands] {
	if {[keylget commStates $comm]!=$mode} {
	    $navwin.$comm configure -state $mode
	    keylset commStates $comm $mode
	    if {!$nestedform} {
		$win.mb.file entryconfigure [keylget menuindex $comm] -state $mode
	    }
	}
    }
}
# Setz modi 
# z.B.
# {{insertonly 0} {nodelete 1} {noupdate 0} {noinsert 1}}
FormEngine instproc setModi newmodi {
    [self] instvar navwin modi sqlconnect usecommands
    if {$sqlconnect==""} {
	return
    }
    set modi $newmodi
    tdom_setdefaults modi {{insertonly 0} {nodelete 0} {noupdate 0} {noinsert 0}}
    set buttons [pack slaves $navwin]
    # vorsicht hacky
    foreach button {delete update} {
	if {[keylget modi no$button]} {
	    if {[lcontain $buttons $navwin.$button]} {
		pack forget $navwin.$button
		set index [lsearch $usecommands $button]
		[self] setFunctionMode disabled $button
		if {$index>=0} {
		    set usecommands [lreplace $usecommands $index $index]
		}
	    } 
	} 
# else {
# 	    if {![lcontain $buttons $navwin.$button]} {
# 		lappend usecommands $button
# 		if {$button=="delete"} {
# 		   pack $navwin.delete -side left -anchor w  -after $navwin.undo
# 		}
# 		if {$button=="update"} {
# 		    pack $navwin.update -side left -anchor w  -bevor $navwin.undo
# 		}
# 	    }
# 	}
    }
}
#
FormEngine instproc stateNeedReload {} {
    [self] instvar navwin sqlconnect
    # Setzt den Hintegrund von Reload Button auf rot
    if {[$sqlconnect hasfunction seek] && [winfo exist $navwin.reload]} {
	$navwin.reload configure -bg red
    }
}
FormEngine instproc reload {} {
    [self] instvar labelwin tupel_count oneobject act_tupel navwin \
	sqlconnect old_tupel modi
    #puts "RELOAD [self] $act_tupel $[self]::old_tupel"
    if {![Object isobject [self]::myDBProxy]} {
	return
    }
    if {![$sqlconnect hasfunction seek]} {
	set act_tupel 0
	set old_tupel -1
    }
    if {[keylget modi insertonly]} {
	set act_tupel 0
	set old_tupel -1
	set tupel_count 0
    } else {
	set tupel_count [[self]::myDBProxy reload]
	if {$tupel_count==-1} {
	    global lan
	    while 1 {
		bgerror $lan(dialog.sqlerror)
		break
	    }
	    ::destroy $navwin
	    set sqlconnect {}
	    [self]::myDBProxy destroy
	    return
	}
	if {!$oneobject} {
	    $labelwin subwidget label configure -text "von $tupel_count"
	}
	if {$tupel_count==0} {
	    set act_tupel 0
	} elseif {$act_tupel>=$tupel_count} {
	    set act_tupel [expr $tupel_count-1]
	}
	$navwin.reload configure -bg #c0c0c0
    }
    # Procedure to add user specific reload
    set old_tupel -1
    [self] reload_form
    [self] check_range
}
FormEngine instproc fill_form {} {
    [self] instvar act_tupel old_tupel tupel_count dialog_container labelwin attrlist oneobject def_values
    #puts "fill_form [self] $old_tupel $act_tupel"
    if {$old_tupel!=-1 && $act_tupel!=$old_tupel} { 
	if {![[self] update]} {
	    set act_tupel $old_tupel
	    return
	}
    }
    if {$act_tupel>=$tupel_count} {
	#Erzeuge ein leeres Prototyp Tupel
	#mit gesetzten default Werten
	foreach attr $attrlist {
	    keylset tupel $attr {}
	}
	$dialog_container set_defaults $tupel
	$dialog_container set_values $def_values
    } else {
	set tupel [[self]::myDBProxy gettupel $act_tupel]
	#puts "SET DEFAULT $tupel"
	# Interface to user api
	[self] filling_form tupel
	$dialog_container set_defaults $tupel
    }
    set old_tupel $act_tupel
}
# Setzt die Verndrung von Form auf Datenbestande 
# Wenn geglckt liefert 1 zurck sonst 0
# durchfhrt eigentlich 2 Operationnen update (modifity) and insert
FormEngine instproc update {} {
    [self] instvar dialog_container old_tupel tupel_count modi
    if {[[self] isinsert]} {
	set ret [[self] insert]
	switch $ret { 
	    0 { return 0 }
	    2 { return 1 }
	}
    } else {
	$dialog_container update_values
	set changes [$dialog_container get_changes]
	#puts "UPDATE [self] $changes"
	# Hier sollte gezeigt werden das die nderungen Ausgelesen wurden
	# wichtig beim save Funktion
	if {[lempty $changes]} { return 1 }
	if {[keylget modi noupdate]} {return 1}
	# Update alter Tupel
	if {[[self] update_check changes]} {
	    if {[[self]::myDBProxy update $changes]==-1} {
		global lan
		while 1 {
		    bgerror $lan(dialog.sqlerror_update)
		    break
		}
		return 0		
	    }
	} else {
	    return 0
	}
    }
    [self] stateNeedReload
    return 1
}
# Setzt den Formular Inhalt als neues Tupel in die DB
# values entspricht is keyed list (aus [$dialog_container get_changes])
# Return Values:
# 0 - can't insert (integrity condition) 
# 1 - insert ok
# 2 - no changes (ignore)
FormEngine instproc insert {} {
    [self] instvar def_values modi nestedform dialog_container notnull_data
    if {[keylget modi noinsert]} {return 0}
    $dialog_container update_values
    set values [$dialog_container get_changes]
    #puts "UPDATE [self] $changes"
    # Hier sollte gezeigt werden das die nderungen Ausgelesen wurden
    # wichtig beim save Funktion
    if {[lempty $values]} { return 2 }
    # Problem keine nderungen der Neu Template
    # Problem wenn alle nderungen der default entsprechen und 
    # es keine zustliche gibt dann ignorieren
    if {[keylequal def_values values]} {
	return 2
    }
    # berprfen der not_null Bedingung
    # alle notnull_data mssen in get_changes vorhanden sein
    set vkeys [keylkeys values]
    foreach nkey $notnull_data {
	if {![lcontain $vkeys $nkey]} {
	    global lan
	    tk_messageBox -title {notnull itegrity volation} -message [format $lan(message.notnull) $nkey] -type ok -icon error
	    return 0
	}
    }
    if {$nestedform} {
	# Mehrfaches Vererbung oder Aggregation
	# id wurde von dem Urvater bereit ermittelt
	if {[[self] insert_check values]} {
	    set newid [[self]::myDBProxy set refvalue]
	    if {[[self]::myDBProxy new $values]=="SQLERROR"} {
		global lan
		while 1 {
		    bgerror $lan(dialog.sqlerror_insert)
		    break
		}
		return 0		
	    }
	} else {
	    return 0
	}
    } else {
	if {[[self] insert_check values]} {
	    if {[set newid [[self]::myDBProxy new $values]]=="SQLERROR"} {
		global lan
		while 1 {
		    bgerror $lan(dialog.sqlerror_insert)
		    break
		}
		return 0		
	    }
	} else {
	    return 0
	}
    }
    #puts "created [self] $newid"
    [self] invokeNotify create $newid 
    return 1
}
# Manual Ausgefhtes Update 
FormEngine instproc updateManual {} {
    [self] instvar dialog_container
    [self] update
    $dialog_container update_defaults
}
# Diese Funktion wird nur durch den angemeldetetn CallBack von
# Vater-FormEngine aufgerufen 
FormEngine instproc createNewTupel {key} {
    #puts "create [self] $key"
    [self] instvar navwin oneobject
    [self]::myDBProxy set refvalue $key
    [self] update
}
FormEngine instproc delete {} {
    [self] instvar old_tupel tupel_count modi 
    # Kein Tupel zum Lschen ist vorhanden
    if {[[self] isinsert] || [keylget modi nodelete]} { 
	return
    }
    if {![[self] delete_check]} return
    if {[[self]::myDBProxy delete]==-1} {
	global lan
	while 1 {
	    bgerror $lan(dialog.sqlerror_delete)
	    break
	}
	return
    }
    [self] stateNeedReload
    #puts "DELETE [self]"
    [self] invokeNotify delete
    [self] after_delete
    [self] next_tupel
}
FormEngine instproc undo {} {
    [self] instvar dialog_container
    $dialog_container undo_changes
}
FormEngine instproc save_tupel {} {
    [self] update
}
# Jetzt nur ein Attribut
# Ausbaufhig
FormEngine instproc orderbyData {} {
    [self] instvar attrlist act_tupel orderby
    set prop1 {{type OComboBox} {name attr} {label attribute} {dropdown 0}}
    set prop2 {{type OCheckButton} {name descendant} {label descendant}}
    keylset prop1 valuelist $attrlist
    if {![info exist orderby]} {
	set orderby {}
    }
    OInputDialog [self]::sortdialog [list $prop1 $prop2] $orderby -returnChanges 0
    if {[set ret [[self]::sortdialog execute]]!=""} {
	set sortlist [keylget ret attr]
	if {[keylget ret descendant]} {
	    lappend sortlist DESC
	}
	[self]::myDBProxy setOrder [list $sortlist]
	set act_tupel 0
	[self] reload
	set orderby $ret
    } else {
	if {$orderby!=""} {
	    [self]::myDBProxy setOrder {}
	    set act_tupel 0
	    [self] reload
	    set orderby ""
	}
    }
    [self]::sortdialog destroy    
}
FormEngine instproc filterData {} {
    [self] instvar navwin checkfilter fwin attrlist dialog_container old_filter 
    global lan
    set fwin [winfo parent $navwin].filter
    if {![winfo exist $fwin]} {
	frame $fwin  -bd 1 -relief ridge
	set checkfilter or
	radiobutton $fwin.checkor -text $lan(navigation.checkor) -value or -variable [self]::checkfilter
	radiobutton $fwin.checkand -text $lan(navigation.checkand) -value and -variable [self]::checkfilter
	tk_optionMenu $fwin.match [self]::filtermatch exact glob regexp
	button $fwin.apply -text $lan(standard.apply) -command "[self] applyFilter"
	button $fwin.drop -text $lan(standard.drop) -command "[self] dropFilter"
	pack $fwin.checkor $fwin.checkand $fwin.match $fwin.apply $fwin.drop -side left -anchor w	
    }
    if {![info exist old_filter]} {
	set old_filter {}
    }
    set fkeys [keylkeys old_filter]
    foreach attr $attrlist {
	if {![lcontain $fkeys $attr]} {
	    keylset old_filter $attr {}
	}
    }
    #  so setzten dass get_changes es erfassen kann
    $dialog_container set_values $old_filter
    pack forget $navwin 
    pack $fwin  -side bottom -anchor s -expand yes -fill x
    
}
FormEngine instproc applyFilter {} {
    [self] instvar dialog_container checkfilter filtermatch navwin fwin old_filter act_tupel old_tupel
    #puts "applyFilter"
    $dialog_container update_values
    set t_changes [$dialog_container get_changes]
    set changes {}
    # Leere Werte ausfiltern 
    foreach key [keylkeys t_changes] {
	if {[keylget t_changes $key]!=""} {
	    keylset changes $key [keylget t_changes $key]
	}
    }
    if {[lempty $changes]} { 
	pack forget $fwin 
	pack $navwin  -side bottom -anchor s -expand yes -fill x
	[self] fill_form
	return
    }
    set old_filter $changes
    [self]::myDBProxy setFilter $changes $checkfilter $filtermatch
    $navwin.find configure -bg blue
    pack forget $fwin 
    pack $navwin  -side bottom -anchor s -expand yes -fill x
    set act_tupel 0
    set old_tupel 0
    [self] reload
}
FormEngine instproc dropFilter {} {
    [self] instvar navwin fwin old_filter navwin
    $navwin.find configure -bg #c0c0c0
    pack forget $fwin 
    pack $navwin  -side bottom -anchor s -expand yes -fill x
    if {$old_filter!=""} {
	set old_filter {}
	[self]::myDBProxy setFilter "" 1 exact
	[self] reload
    } else {
	[self] fill_form
    }
}
FormEngine instproc check_range {} {
    [self] instvar act_tupel old_tupel tupel_count modi oneobject labelwin
    if {![[self] position_check $act_tupel]} {
	set act_tupel $old_tupel
	return
    }
    if {$act_tupel<0 || ($act_tupel>[expr $tupel_count - [keylget modi noinsert]] && $tupel_count!=0)} {
	set act_tupel $old_tupel
    } else {
	set my_old $old_tupel
	[self] fill_form
	if {$act_tupel==0} {
	    [self] setFunctionMode disabled {first prev}
	} elseif {$my_old==0 || $my_old==-1} {
	    [self] setFunctionMode normal {first prev}
	}
	if {$act_tupel==$tupel_count-1 && [keylget modi noinsert]} {
	    [self] setFunctionMode disabled {last next}
	} elseif {($my_old==$tupel_count-1 && [keylget modi noinsert]) || $my_old==-1} {
	    [self] setFunctionMode normal {last next}
	}
	if {$act_tupel>=$tupel_count} {
	    [self] setFunctionMode disabled delete
	} elseif {$my_old>=$tupel_count || $my_old==-1} {
	    [self] setFunctionMode normal delete
	}
    }
    if {!$oneobject && ![keylget modi insertonly]} {
	$labelwin subwidget entry delete 0 end
	$labelwin subwidget entry insert 0 [expr $act_tupel +1]
    }
}
# Erzeugt neues Tupel aus Eingabemaske und setzt
# die Eingabemaske fr die neuste Eingabe
# Der Neue Tupel wird nicht die Tupelzahl erhhen
# SQL Spezifisch
FormEngine instproc handle_newtupel {} {
    if {[[self] insert]==1} {
	[self] fill_form
	[self] stateNeedReload
    }
}
FormEngine instproc next_tupel {} {
    [self] instvar act_tupel
    if {[[self] isinsert]} {
	[self] handle_newtupel
    } else {
	incr act_tupel
	[self] check_range
    }
}
FormEngine instproc prev {} {
    [self] instvar act_tupel
    incr act_tupel -1
    [self] check_range
}
FormEngine instproc first {} {
    [self] instvar act_tupel
    set act_tupel 0
    [self] check_range
}
FormEngine instproc last {} {
    [self] instvar act_tupel old_tupel tupel_count modi
    if {[[self] isinsert]} {
	[self] handle_newtupel
    } else {
	if {$old_tupel==$tupel_count-1} {
	    set act_tupel [expr $tupel_count - [keylget modi noinsert]]
	} else {
	    set act_tupel [expr $tupel_count - 1]
	}
	[self] check_range
    }
}
FormEngine instproc newtupel {} {
    [self] instvar act_tupel old_tupel tupel_count modi
    if {[[self] isinsert]} {
	[self] handle_newtupel
    } else {
	set act_tupel $tupel_count
	[self] check_range
    }
}
# Abfrage ob gerede ein neuer Tupel eingegeben wird
FormEngine instproc isinsert {} {
    [self] instvar old_tupel tupel_count
    return [expr $old_tupel>=$tupel_count]
}
FormEngine instproc setPosFromEntry {} {
    [self] instvar labelwin act_tupel
    set pos [$labelwin subwidget entry get]
    if {[catch {expr $pos+1}]} {
	$labelwin subwidget entry delete 0 end
	$labelwin subwidget entry insert 0 [expr $act_tupel +1]
    } else {
	set act_tupel [expr $pos-1]
	[self] check_range
    }
}
# Forn Lschen SQL Querry beenden
FormEngine instproc destroy {} {
    [self] instvar win linkedForm oneobject sqlconnect
    global globalexit
    if {!$globalexit} {
	if {$sqlconnect!=""} {
	    # wenn es etwas nicht stimmt dann nicht aufhngen
	    if {[catch {[self] update} er]!=0} {
		global errorInfo
		puts "update Error by [self] $er $errorInfo"
	    }
	}
	#puts "destroying FormEngine"
	if {[info exist linkedForm] && [FormEngine deepinstances $linkedForm]!=""} {
	$linkedForm clearNotify create "[self] createNewTupel"
	    # Hier kann man auch ein Kaskadierendes Delete Aufbauen"
	    if {$oneobject} {
		$linkedForm clearNotify delete "[self] delete"
	    }
	}
	# zuerst die children lschen (sie brauchen die Windows noch fr update)
	next
	#     foreach child [[self] info children] {
	# 	if {[$child info class]=="::FrameContainer"} {
	# 	    $child destroy
	# 	}
	#     }
	::destroy $win
	#    next
    }
}
# Prozeduren fr die Behandlung der Formen als EditWidget
FormEngine instproc setvalue {value} {
}
FormEngine instproc updatevalue {} {
    [self] instvar linkedForm
    #puts "updatevalue [self]"
    # Ist der Formular gelinkt
    if {[info exist linkedForm] && [FormEngine deepinstances $linkedForm]!=""} {
	# wenn Update beim Create Tupel, dann ignoriere
	# dafr gibt es Notify createNewTupel
	if {[$linkedForm isinsert]} {
	    return
	}
    }
    [self] update
}
FormEngine instproc getvalue {} {
    [self] instvar value
    return $value
}
# Vorsicht!
# setdefault wird auch beim undo_changes ausgefhrt
FormEngine instproc setdefault {} {
    [self] instvar default value act_tupel tupel_count old_tupel navwin oneobject sqlconnect emptyLinkState
    set value $default
    #puts "setdefault [self] VAL=$value"
    set act_tupel 0
    set old_tupel -1
    # value=="" soll als neues Tupel implementiert werden
    #if {$value==""} return
    if {$default==""} {
	set tupel_count 0
	set emptyLinkState 1
	[self] fill_form
	[self] setFunctionMode disabled {first prev last next delete 
	    reload find orderby update newtupel}
    } else {
	if {$emptyLinkState} {
	    [self] setFunctionMode normal {first prev last next delete 
		reload find orderby update newtupel}
	}
	set emptyLinkState 0
	[self]::myDBProxy set refvalue $default
	[self] reload
    }
}
# Prozeduren, die die Introspektion info erweitern
FormEngine proc deepinstances {{inst {}}} {
    if {$inst==""} {
	return [[self] searchInstances ::FormEngine]
    } else {
	return [[self] searchForInstance ::FormEngine $inst]
    }
}
FormEngine proc searchInstances {class} {
    set ret [$class info instances] 
    foreach subclass [$class info subclass] {
	set ret [concat $ret [[self] searchInstances $subclass]]
    }
    return $ret
}
FormEngine proc searchForInstance {class inst} {
    set ret [$class info instances $inst] 
    if {$ret!=""} { 
	return $ret
    }
    foreach subclass [$class info subclass] {
	set ret [[self] searchForInstance $subclass $inst]
	if {$ret!=""} { 
	    return $ret
	}
    }
    return ""
}
# Diese Prozedur dient zu anbindung von Makros 
# Dabei wird nicht ein Objekt der Klasse FormEngine gebildet
# sondern wenn mglich eins von seinen Subclass
# mit dem Namen [namespace tail $sclass]==$form_name
FormEngine proc createPerName {form_name name args} {
    set fclass FormEngine
    foreach sclass [FormEngine info subclass] {
	if {[namespace tail $sclass]==$form_name} {
	    set fclass $sclass
	    break
	}
    }
    return [eval $fclass $name $args]
}

Class FormLink  -superclass InputWidget -parameter {
    form_node
    sqlconnect 
    data
    label
    {refattribute {}}
    reftable
    objlinks
}
FormLink metadata add attributes
# objlinks hat Aufbau
# {{form refattribute label} {form refattribute label} ...}
FormLink metadata attributes {label form refattribute data reftable reftable}
#
# Realiseirung von Form Navigation (FormLink)
# Das Eigentliche Form wird in das Objekt Aggregiert  
# die nderungen werden durch 
FormLink instproc map {window} {
    [self] instvar win label reftable objlinks
    set win $window
    if {![info exist reftable]} {
	button $win -text $label -command "[self] executeForm"
    } else {
	menubutton $win -text $label -menu $win.menu -relief raised
	menu $win.menu -tearoff 0
	$win.menu add command -label $label -command "[self] executeForm"
	foreach link $objlinks {
	    $win.menu add command -label [lindex $link 2] -command "[self] executeForm [lindex $link 0]"
	}

    }
    return $win
}
FormLink instproc executeForm {{obj {}}} {
    [self] instvar form_node sqlconnect data refattribute \
	formengine objformengine objlinks reftable
    # Normalles 1:n Link
    if {$obj==""} {
	if {![info exist formengine] || \
		$formengine=="" || \
		[FormEngine deepinstances ::$formengine]==""} {
	    # suche den Vater FormEngine
	    # FormLink soll der Kind oder Enkelkind .... davon sein
	    set pf_engine [FormEngine searchParentFormEngine [self]]
	    set formengine [FormEngine createPerName \
				[$form_node getAttribute name] \
				[Object autoname formengine] \
				-form_node $form_node \
				-sqlconnect $sqlconnect -data $data \
				-refattribute $refattribute -linkedForm $pf_engine 
			   ]
	    $formengine map
	    [self] setdefault
	}
    } else {
	if {![info exist objformengine($obj)] || \
	    $objformengine($obj)=="" || \
	    [FormEngine deepinstances ::$objformengine($obj)]==""} {
	    set pf_engine [FormEngine searchParentFormEngine [self]]
	    set refa [lindex [lindex $objlinks [lsearch -glob $objlinks $obj*]] 1]
	    set oform_node [tdom_findpername $form_node form $obj]
	    set  objformengine($obj) \
		[FormEngine createPerName \
		     [$oform_node getAttribute name] \
		     [Object autoname formengine] \
		     -form_node $oform_node \
		     -sqlconnect $sqlconnect -data $data \
		     -refattribute $refattribute -linkedForm $pf_engine \
		     -reftableLink [list $reftable $refa] \
		     -modi {{nodelete 1} {noinsert 1}}] 
	    $objformengine($obj) map
	    [self] setdefault
	}
    }
} 

# Prozeduren fr die Behandlung der FormLink als EditWidget
FormLink instproc setvalue {value} {
}
FormLink instproc updatevalue {} {
    [self] instvar formengine objformengine
    #puts "updatevalue [self] [self class]"
    if {[info exist formengine] &&  $formengine!="" &&  [FormEngine deepinstances ::$formengine]!=""} {
	$formengine updatevalue
    }
    if {[array exist objformengine]} {
	foreach obj [array names objformengine] {
	    if {$objformengine($obj)!="" &&  [FormEngine deepinstances ::$objformengine($obj)]!=""} {
		$objformengine($obj) updatevalue
	    }	
	}
    }
}
FormLink instproc getvalue {} {
    [self] instvar value
    return $value
}
FormLink instproc setdefault {} {
    [self] instvar default value formengine objformengine sqlconnect
    #puts "setdefault [self] [self class]"
    if {$sqlconnect==""} return
    set value $default
    if {[info exist formengine] &&  $formengine!="" &&  [FormEngine deepinstances ::$formengine]!=""} {
	$formengine set default $default
	$formengine setdefault
    }
    if {[array exist objformengine]} {
	foreach obj [array names objformengine] {
	    if {$objformengine($obj)!="" &&  [FormEngine deepinstances ::$objformengine($obj)]!=""} {
		$objformengine($obj) default $default
		$objformengine($obj) setdefault
	    }	
	}
    }
}
#
# Die Vererbung wird durch ein NoteBook Widget wiedergeben.
# Dabei werden die unterschiedliche Argumente des 
# Kindobjekt als Unterformulare (Eigebetette Formulare) gezeigt
#
# Das Objekt besitzt entweder aggregiertes ONoteBookConteiner Objekt mit
# einzelnen Unterformularen(FormEngine)  (notebook) oder eine aggregierte 
# FormEngine als (form)
#
# ! Die Objekte der Klasse knnen sich selber nicht kontruieren
Class InheritForm  -superclass InputWidget 
InheritForm metadata add attributes
InheritForm metadata attributes {data}
InheritForm instproc map {window} {
    [self] instvar win
    set win $window
    if {[[self] info children notebook]!=""} {
	[self]::notebook map $window
    } else {
	[self]::form map $window
    }
    return $win
}
InheritForm instproc setvalue {value} {
}
InheritForm instproc getAllChildren {} {
    if {[[self] info children notebook]!=""} {
	return  [[self]::notebook info children]
    } else {
	return [self]::form
    }
}
InheritForm instproc updatevalue {} {
    [self] instvar data
    foreach child [[self] getAllChildren] {
	    $child updatevalue
    }
}
InheritForm instproc getvalue {} {
    [self] instvar value
    return $value
}
InheritForm instproc setdefault {} {
    [self] instvar default value data 
    set value $default
    # Eigentlich ein Neues Template
    set foundchild 0
    foreach child [[self] getAllChildren] {
	#puts "seting $child"
	$child set default $default
	$child setdefault
	#[self]::notebook::${echild}::myDBProxy set refvalue $default
	#[self]::notebook::${echild} reload
	# Schalte auf das NoteBook wenn Kind gefunden
	if {!$foundchild && [[self] info children notebook]!=""} {
	    if {[$child set tupel_count]>0} {
		[self]::notebook raise [string tolower [namespace tail $child]]
		set foundchild 1
	    }
	}
	
    }
}
# Ein Widget fr Daertellung der Referenzen
# ab Version 0.12 kann es als FormLink benutzt werden
Class OSqlReferenceList  -superclass {OMultiColumnList FormLink} -parameter {
    sqlquery
    sqlconnect
    {notnull 0}
    {showfirst 1}
    value
    {default {}}
    data
} 
OSqlReferenceList metadata add description
OSqlReferenceList metadata description {
Humankonforme Reprsentation der Fremdschlssel als 
mehrspaltige Auswahlliste
hat kein Attribut valuelist
setvalue und getvalue behandeln den Index statt konkreten Werten.
Hat kein setdefault value
Vorsicht Parameter sqlconnect mu initialisiert werden
auch notnull gehrt nicht zur Klassenvarible attributes
(Stadart Datenverkunft Parameter)
Parameter:  state valuelist columns columnslabel columnswidth height notnull
sqlconnect sqlquery
}
OSqlReferenceList metadata add attributes
OSqlReferenceList metadata attributes {state label columns columnslabel height columnswidth sqlquery showfirst dropdown form}
OSqlReferenceList instproc map window {
    [self] instvar win form_node refattribute
    next
    if {[info exist form_node]} {
 	bind $win <Double-Button-1> "[self] executeForm"
	bind [$win subwidget frame] <Double-Button-1> "[self] executeForm"
	set refattribute [$form_node getAttribute primarykey]
    }
    refresh
}
OSqlReferenceList instproc refresh {} {
    [self] instvar sqlconnect sqlquery notnull columns indexlist showfirst win
    if {$sqlconnect==""} return
    set query [$sqlconnect query $sqlquery]
    if {$query==""} {
	global lan
	while 1 {
	    bgerror "$lan(dialog.sqlerror_reference)"
	    break
	}
	$win configure -bg red
	set sqlconnect {}
	return
    }
    if {!$notnull} {
	for {set x 0} {$x<$columns} {incr x} {
	    lappend nulllist NULL
	}
	lappend valuelist $nulllist
	lappend indexlist NULL
    }
    for {set x [$query rows]} {$x>0} {incr x -1} {
	# is der Eigentliche Fremschlssel ausgeblendet
	set row  [$query fetch]
	lappend indexlist [lindex $row 0]
	if {$showfirst} {
	    lappend valuelist $row
	} else {
	    lappend valuelist [lrange $row 1 end]
	}
    }
    $query destroy
    [self] setvaluelist $valuelist
}
OSqlReferenceList instproc getvalue {} {
    [self] instvar value win
    [self] updatevalue
    return $value
}
OSqlReferenceList instproc updatevalue {} {
    [self] instvar value win indexlist  form_node
    if {[info exist indexlist]} {
	set index [[self] getindex]
	if {$index!=""} {
	    set value [lindex $indexlist $index]
	}
    }
    if {[info exist form_node]} {
	next
    }
    #puts "upadet sqllist $value"
}
OSqlReferenceList instproc setvalue {newv} {
    [self] instvar win value indexlist  form_node
    global lan
    set value $newv
    # Lerres Wert (z.B. neuer Tulep 
    if {[info exist indexlist]} {
	if {$value==""} {
	    [self] setindex {}
	    return
	}
	set index [lsearch $indexlist $value]
	if {$index==-1} {
	    tk_messageBox -message $lan(dialog.noprimarykey) -type ok -icon warning
	    return
	}
	[self] setindex $index
	[self] yview $index
    }
    if {[info exist form_node]} {
 	next
    }
}
OSqlReferenceList instproc setdefault {} {
    #puts "setdefault [self] [self class]"
    [self] instvar win default form_node
    [self] setvalue $default
    if {[info exist form_node]} {
 	next
    }
}
